# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Stochastic weight averaging callback

Modified from:
    lightning.pytorch.callbacks.StochasticWeightAveraging
"""

from typing import Any, Callable, List, Optional, Union, cast

import lightning.pytorch as pl
import torch
from lightning.fabric.utilities.types import LRScheduler
from lightning.pytorch.callbacks import StochasticWeightAveraging as _StochasticWeightAveraging
from lightning.pytorch.utilities.rank_zero import rank_zero_info, rank_zero_warn
from lightning.pytorch.utilities.types import LRSchedulerConfig
from torch import Tensor
from torch.optim.swa_utils import SWALR

_AVG_FN = Callable[[Tensor, Tensor, Tensor], Tensor]
_DEFAULT_DEVICE = torch.device("cpu")


class StochasticWeightAveraging(_StochasticWeightAveraging):
    """Stochastic weight averaging callback

    Modified from:
        lightning.pytorch.callbacks.StochasticWeightAveraging
    """

    def __init__(
        self,
        swa_lrs: Union[float, List[float]],
        swa_epoch_start: Union[int, float] = 0.8,
        annealing_epochs: int = 10,
        annealing_strategy: str = "cos",
        avg_fn: Optional[_AVG_FN] = None,
        device: Optional[Union[torch.device, str]] = _DEFAULT_DEVICE,
    ):
        r"""Implements the Stochastic Weight Averaging (SWA) Callback to average a model.

        Stochastic Weight Averaging was proposed in ``Averaging Weights Leads to
        Wider Optima and Better Generalization`` by Pavel Izmailov, Dmitrii
        Podoprikhin, Timur Garipov, Dmitry Vetrov and Andrew Gordon Wilson
        (UAI 2018).

        This documentation is highly inspired by PyTorch's work on SWA.
        The callback arguments follow the scheme defined in PyTorch's ``swa_utils`` package.

        For a SWA explanation, please take a look
        `here <https://pytorch.org/blog/pytorch-1.6-now-includes-stochastic-weight-averaging>`_.

        .. warning::  This is an :ref:`experimental <versioning:Experimental API>` feature.

        .. warning:: ``StochasticWeightAveraging`` is currently not supported for multiple
            optimizers/schedulers.

        .. warning:: ``StochasticWeightAveraging`` is currently only supported on every epoch.

        Arguments:
            swa_lrs: The SWA learning rate to use:

                - ``float``. Use this value for all parameter groups of the optimizer.
                - ``List[float]``. A list values for each parameter group of the optimizer.

            swa_epoch_start: If provided as int, the procedure will start from
                the ``swa_epoch_start``-th epoch. If provided as float between 0 and 1,
                the procedure will start from ``int(swa_epoch_start * max_epochs)`` epoch

            annealing_epochs: number of epochs in the annealing phase (default: 10)

            annealing_strategy: Specifies the annealing strategy (default: "cos"):

                - ``"cos"``. For cosine annealing.
                - ``"linear"`` For linear annealing

            avg_fn: the averaging function used to update the parameters;
                the function must take in the current value of the
                :class:`AveragedModel` parameter, the current value of :attr:`model`
                parameter and the number of models already averaged; if None,
                equally weighted average is used (default: ``None``)

            device: if provided, the averaged model will be stored on the ``device``.
                When None is provided, it will infer the `device` from ``pl_module``.
                (default: ``"cpu"``)

        """
        # Add this so we can use iterative datapipe
        self._train_batches = 0

        super()._init_(
            swa_lrs,
            swa_epoch_start,
            annealing_epochs,
            annealing_strategy,
            avg_fn,
            device,
        )

    def on_train_epoch_start(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        """Run at start of each train epoch"""
        if (not self._initialized) and (self.swa_start <= trainer.current_epoch <= self.swa_end):
            self._initialized = True

            # move average model to request device.
            assert self._average_model is not None
            self._average_model = self._average_model.to(self._device or pl_module.device)

            optimizer = trainer.optimizers[0]
            if isinstance(self._swa_lrs, float):
                self._swa_lrs = [self._swa_lrs] * len(optimizer.param_groups)

            for lr, group in zip(self._swa_lrs, optimizer.param_groups, strict=True):
                group["initial_lr"] = lr

            assert trainer.max_epochs is not None
            self._swa_scheduler = cast(
                LRScheduler,
                SWALR(
                    optimizer,
                    swa_lr=self._swa_lrs,  # type: ignore[arg-type]
                    anneal_epochs=self._annealing_epochs,
                    anneal_strategy=self._annealing_strategy,
                    last_epoch=trainer.max_epochs if self._annealing_strategy == "cos" else -1,
                ),
            )
            if self._scheduler_state is not None:
                # Restore scheduler state from checkpoint
                self._swa_scheduler.load_state_dict(self._scheduler_state)
            elif trainer.current_epoch != self.swa_start:
                # Log a warning if we're initializing after start without any checkpoint data,
                # as behaviour will be different compared to having checkpoint data.
                rank_zero_warn(
                    "SWA is initializing after swa_start without any checkpoint data. "
                    "This may be caused by loading a checkpoint from an older version of PyTorch"
                    " Lightning."
                )

            # We assert that there is only one optimizer on fit start
            default_scheduler_cfg = LRSchedulerConfig(self._swa_scheduler)
            assert default_scheduler_cfg.interval == "epoch"
            assert default_scheduler_cfg.frequency == 1

            if trainer.lr_scheduler_configs:
                scheduler_cfg = trainer.lr_scheduler_configs[0]
                if scheduler_cfg.interval != "epoch" or scheduler_cfg.frequency != 1:
                    rank_zero_warn(
                        f"SWA is currently only supported every epoch. Found {scheduler_cfg}"
                    )
                rank_zero_info(
                    f"Swapping scheduler `{scheduler_cfg.scheduler.__class__.__name__}`"
                    f" for `{self._swa_scheduler.__class__.__name__}`"
                )
                trainer.lr_scheduler_configs[0] = default_scheduler_cfg
            else:
                trainer.lr_scheduler_configs.append(default_scheduler_cfg)

            if self.n_averaged is None:
                self.n_averaged = torch.tensor(
                    self._init_n_averaged, dtype=torch.long, device=pl_module.device
                )

        if (self.swa_start <= trainer.current_epoch <= self.swa_end) and (
            trainer.current_epoch > self._latest_update_epoch
        ):
            assert self.n_averaged is not None
            assert self._average_model is not None
            self.update_parameters(self._average_model, pl_module, self.n_averaged, self._avg_fn)
            self._latest_update_epoch = trainer.current_epoch

        # Note: No > here in case the callback is saved with the model and training continues
        if trainer.current_epoch == self.swa_end + 1:
            # Transfer weights from average model to pl_module
            assert self._average_model is not None
            self.transfer_weights(self._average_model, pl_module)

            # Reset BatchNorm for update
            self.reset_batch_norm_and_save_state(pl_module)

            # There is no need to perform either backward or optimizer.step as we are
            # performing only one pass over the train data-loader to compute activation statistics
            # Therefore, we will virtually increase the number of training batches by 1 and
            # skip backward.
            trainer.fit_loop.max_batches += 1
            trainer.fit_loop._skip_backward = True
            self._accumulate_grad_batches = trainer.accumulate_grad_batches
            trainer.accumulate_grad_batches = self._train_batches

    def on_train_epoch_end(self, trainer: "pl.Trainer", *args: Any) -> None:
        """Run at end of each train epoch"""
        if trainer.current_epoch == 0:
            self._train_batches = trainer.global_step
        trainer.fit_loop._skip_backward = False



import lightning


def test_model_trainer_fit(model, sample_datamodule):
    trainer = lightning.pytorch.trainer.trainer.Trainer(fast_dev_run=True)
    trainer.fit(model=model, datamodule=sample_datamodule)



"""Simple model which only uses outputs of PVNet for all GSPs"""

from typing import Optional

import numpy as np
import pvnet
import torch
import torch.nn.functional as F
from pvnet.models.multimodal.linear_networks.basic_blocks import AbstractLinearNetwork
from pvnet.models.multimodal.linear_networks.networks import DefaultFCNet
from pvnet.optimizers import AbstractOptimizer
from torch import nn

from pvnet_summation.models.base_model import BaseModel

_default_optimizer = pvnet.optimizers.Adam()


class Model(BaseModel):
    """Neural network which combines GSP predictions from PVNet"""

    name = "pvnet_summation_model"

    def __init__(
        self,
        model_name: str,
        model_version: Optional[str],
        output_quantiles: Optional[list[float]] = None,
        output_network: AbstractLinearNetwork = DefaultFCNet,
        output_network_kwargs: Optional[dict] = None,
        relative_scale_pvnet_outputs: bool = False,
        predict_difference_from_sum: bool = False,
        optimizer: AbstractOptimizer = _default_optimizer,
    ):
        """Neural network which combines GSP predictions from PVNet

        Args:
            model_name: Model path either locally or on huggingface.
            model_version: Model version if using huggingface. Set to None if using local.
            output_quantiles: A list of float (0.0, 1.0) quantiles to predict values for. If set to
                None the output is a single value.
            output_network: Pytorch Module class used to combine the 1D features to produce the
                forecast.
            output_network_kwargs: Dictionary of optional kwargs for the `output_network` module.
            relative_scale_pvnet_outputs: If true, the PVNet predictions are scaled by a factor
                which is proportional to their capacities.
            predict_difference_from_sum: Whether to use the sum of GSPs as an estimate for the
                national sum and train the model to correct this estimate. Otherwise the model tries
                to learn the national sum from the PVNet outputs directly.
            optimizer (AbstractOptimizer): Optimizer
        """

        super().__init__(model_name, model_version, optimizer, output_quantiles)

        self.relative_scale_pvnet_outputs = relative_scale_pvnet_outputs
        self.predict_difference_from_sum = predict_difference_from_sum

        if output_network_kwargs is None:
            output_network_kwargs = dict()

        self.model = output_network(
            in_features=np.product(self.pvnet_output_shape),
            out_features=self.num_output_features,
            **output_network_kwargs,
        )

        # Add linear layer if predicting difference from sum
        # This allows difference to be positive or negative
        if predict_difference_from_sum:
            self.model = nn.Sequential(
                self.model, nn.Linear(self.num_output_features, self.num_output_features)
            )

        self.save_hyperparameters()

    def forward(self, x):
        """Run model forward"""

        if "pvnet_outputs" not in x:
            x["pvnet_outputs"] = self.predict_pvnet_batch(x["pvnet_inputs"])

        if self.relative_scale_pvnet_outputs:
            if self.pvnet_model.use_quantile_regression:
                eff_cap = x["effective_capacity"].unsqueeze(-1)
            else:
                eff_cap = x["effective_capacity"]

            # Multiply by (effective capacity / 100) since the capacities are roughly of magnitude
            # of 100 MW. We still want the inputs to the network to be order of magnitude 1.
            x_in = x["pvnet_outputs"] * (eff_cap / 100)
        else:
            x_in = x["pvnet_outputs"]

        x_in = torch.flatten(x_in, start_dim=1)
        out = self.model(x_in)

        if self.use_quantile_regression:
            # Shape: batch_size, seq_length * num_quantiles
            out = out.reshape(out.shape[0], self.forecast_len_30, len(self.output_quantiles))

        if self.predict_difference_from_sum:
            gsp_sum = self.sum_of_gsps(x)

            if self.use_quantile_regression:
                gsp_sum = gsp_sum.unsqueeze(-1)

            out = F.leaky_relu(gsp_sum + out)

        return out



import os

import pytest
import pandas as pd
import numpy as np
import xarray as xr
import torch
import math
import glob
import tempfile
from pvnet_summation.models.model import Model


from ocf_datapipes.utils.consts import BatchKey
from datetime import timedelta

from pvnet_summation.data.datamodule import DataModule


@pytest.fixture()
def sample_data():
    # Copy small batches to fake 317 GSPs in each
    with tempfile.TemporaryDirectory() as tmpdirname:
        os.makedirs(f"{tmpdirname}/train")
        os.makedirs(f"{tmpdirname}/val")

        # Grab times from batch to make national output zarr
        times = []

        file_n = 0
        for file in glob.glob("tests/data/sample_batches/train/*.pt"):
            batch = torch.load(file)

            this_batch = {}
            for i in range(batch[BatchKey.gsp_time_utc].shape[0]):
                # Duplicate sample to fake 317 GSPs
                for key in batch.keys():
                    if isinstance(batch[key], torch.Tensor):
                        n_dims = len(batch[key].shape)
                        repeats = (317,) + tuple(1 for dim in range(n_dims - 1))
                        this_batch[key] = batch[key][i : i + 1].repeat(repeats)[:317]
                    else:
                        this_batch[key] = batch[key]

                # Save fopr both train and val
                torch.save(this_batch, f"{tmpdirname}/train/{file_n:06}.pt")
                torch.save(this_batch, f"{tmpdirname}/val/{file_n:06}.pt")

                file_n += 1

                times += [batch[BatchKey.gsp_time_utc][i].numpy().astype("datetime64[s]")]

        times = np.unique(np.sort(np.concatenate(times)))

        da_output = xr.DataArray(
            data=np.random.uniform(size=(len(times), 1)),
            dims=["datetime_gmt", "gsp_id"],
            coords=dict(
                datetime_gmt=times,
                gsp_id=[0],
            ),
        )

        da_cap = xr.DataArray(
            data=np.ones((len(times), 1)),
            dims=["datetime_gmt", "gsp_id"],
            coords=dict(
                datetime_gmt=times,
                gsp_id=[0],
            ),
        )

        ds = xr.Dataset(
            data_vars=dict(
                generation_mw=da_output,
                installedcapacity_mwp=da_cap,
                capacity_mwp=da_cap,
            ),
        )

        ds.to_zarr(f"{tmpdirname}/gsp.zarr")

        yield tmpdirname, f"{tmpdirname}/gsp.zarr"


@pytest.fixture()
def sample_datamodule(sample_data):
    batch_dir, gsp_zarr_dir = sample_data

    dm = DataModule(
        batch_dir=batch_dir,
        gsp_zarr_path=gsp_zarr_dir,
        batch_size=2,
        num_workers=0,
        prefetch_factor=2,
    )

    return dm


@pytest.fixture()
def sample_batch(sample_datamodule):
    batch = next(iter(sample_datamodule.train_dataloader()))
    return batch


@pytest.fixture()
def model_kwargs():
    kwargs = dict(
        model_name="openclimatefix/pvnet_v2",
        model_version="898630f3f8cd4e8506525d813dd61c6d8de86144",
    )
    return kwargs


@pytest.fixture()
def model(model_kwargs):
    model = Model(**model_kwargs)
    return model


@pytest.fixture()
def quantile_model(model_kwargs):
    model = Model(output_quantiles=[0.1, 0.5, 0.9], **model_kwargs)
    return model



"""Training"""
import os
from typing import Optional

import hydra
import torch
from lightning.pytorch import (
    Callback,
    LightningDataModule,
    LightningModule,
    Trainer,
    seed_everything,
)
from lightning.pytorch.callbacks import ModelCheckpoint
from lightning.pytorch.loggers import Logger
from lightning.pytorch.loggers.wandb import WandbLogger
from omegaconf import DictConfig, OmegaConf
from pvnet import utils
from tqdm import tqdm

from pvnet_summation.data.datamodule import PVNetPresavedDataModule

log = utils.get_logger(__name__)

torch.set_default_dtype(torch.float32)


def resolve_monitor_loss(output_quantiles):
    """Return the desired metric to monitor based on whether quantile regression is being used.

    The adds the option to use something like:
        monitor: "${resolve_monitor_loss:${model.output_quantiles}}"

    in early stopping and model checkpoint callbacks so the callbacks config does not need to be
    modified depending on whether quantile regression is being used or not.
    """
    if output_quantiles is None:
        return "MAE/val"
    else:
        return "quantile_loss/val"


OmegaConf.register_new_resolver("resolve_monitor_loss", resolve_monitor_loss)


def train(config: DictConfig) -> Optional[float]:
    """Contains training pipeline.

    Instantiates all PyTorch Lightning objects from config.

    Args:
        config (DictConfig): Configuration composed by Hydra.

    Returns:
        Optional[float]: Metric score for hyperparameter optimization.
    """

    # Set seed for random number generators in pytorch, numpy and python.random
    if "seed" in config:
        seed_everything(config.seed, workers=True)

    # Init lightning datamodule
    log.info(f"Instantiating datamodule <{config.datamodule._target_}>")
    datamodule: LightningDataModule = hydra.utils.instantiate(config.datamodule)

    # Init lightning model
    log.info(f"Instantiating model <{config.model._target_}>")
    model: LightningModule = hydra.utils.instantiate(config.model)

    # Presave batches
    if config.get("presave_pvnet_outputs", False):
        save_dir = (
            f"{config.datamodule.batch_dir}/"
            f"{config.model.model_name}/"
            f"{config.model.model_version}"
        )

        if os.path.isdir(save_dir):
            log.info(
                f"PVNet output directory already exists: {save_dir}\n"
                "Skipping saving new outputs. The existing saved outputs will be loaded."
            )

        else:
            log.info(f"Saving PVNet outputs to {save_dir}")

            os.makedirs(f"{save_dir}/train")
            os.makedirs(f"{save_dir}/val")

            # Set batch size to None so batching is skipped
            datamodule.batch_size = None

            for dataloader_func, split in [
                (datamodule.train_dataloader, "train"),
                (datamodule.val_dataloader, "val"),
            ]:
                log.info(f"Saving {split} outputs")
                dataloader = dataloader_func(shuffle=False, add_filename=True)

                for concurrent_sample_dict in tqdm(dataloader):
                    # Run though model and remove
                    pvnet_out = model.predict_pvnet_batch([concurrent_sample_dict["pvnet_inputs"]])[
                        0
                    ]
                    del concurrent_sample_dict["pvnet_inputs"]
                    concurrent_sample_dict["pvnet_outputs"] = pvnet_out

                    # Save pvnet prediction sample
                    filepath = concurrent_sample_dict.pop("filepath")
                    sample_rel_path = filepath.removeprefix(config.datamodule.batch_dir)
                    sample_path = f"{save_dir}{sample_rel_path}"
                    torch.save(concurrent_sample_dict, sample_path)

        datamodule = PVNetPresavedDataModule(
            batch_dir=save_dir,
            batch_size=config.datamodule.batch_size,
            num_workers=config.datamodule.num_workers,
            prefetch_factor=config.datamodule.prefetch_factor,
        )

    # Init lightning loggers
    loggers: list[Logger] = []
    if "logger" in config:
        for _, lg_conf in config.logger.items():
            if "_target_" in lg_conf:
                log.info(f"Instantiating logger <{lg_conf._target_}>")
                loggers.append(hydra.utils.instantiate(lg_conf))

    # Init lightning callbacks
    callbacks: list[Callback] = []
    if "callbacks" in config:
        for _, cb_conf in config.callbacks.items():
            if "_target_" in cb_conf:
                log.info(f"Instantiating callback <{cb_conf._target_}>")
                callbacks.append(hydra.utils.instantiate(cb_conf))

    # Align the wandb id with the checkpoint path
    # - only works if wandb logger and model checkpoint used
    # - this makes it easy to push the model to huggingface
    use_wandb_logger = False
    for logger in loggers:
        log.info(f"{logger}")
        if isinstance(logger, WandbLogger):
            use_wandb_logger = True
            wandb_logger = logger
            break

    if use_wandb_logger:
        for callback in callbacks:
            log.info(f"{callback}")
            if isinstance(callback, ModelCheckpoint):
                callback.dirpath = "/".join(
                    callback.dirpath.split("/")[:-1] + [wandb_logger.version]
                )
                # Also save model config here - this makes for easy model push to huggingface
                os.makedirs(callback.dirpath, exist_ok=True)
                OmegaConf.save(config.model, f"{callback.dirpath}/model_config.yaml")
                break

    trainer: Trainer = hydra.utils.instantiate(
        config.trainer,
        logger=loggers,
        _convert_="partial",
        callbacks=callbacks,
    )

    # Train the model completely
    trainer.fit(model=model, datamodule=datamodule)

    # Validate after end - useful if using stochastic weight averaging
    trainer.validate(model=model, datamodule=datamodule)

    # Make sure everything closed properly
    log.info("Finalizing!")
    utils.finish(
        config=config,
        model=model,
        datamodule=datamodule,
        trainer=trainer,
        callbacks=callbacks,
        loggers=loggers,
    )

    # Print path to best checkpoint
    log.info(f"Best checkpoint path:\n{trainer.checkpoint_callback.best_model_path}")

    # Return metric score for hyperparameter optimization
    optimized_metric = config.get("optimized_metric")
    if optimized_metric:
        return trainer.callback_metrics[optimized_metric]



"""Utils"""

import matplotlib.pyplot as plt
import pandas as pd
import pylab


def plot_forecasts(y, y_hat, times, batch_idx=None, quantiles=None, y_sum=None):
    """Plot a batch of data and the forecast from that batch"""

    times_utc = times.cpu().numpy().squeeze().astype("datetime64[s]")
    times_utc = [pd.to_datetime(t) for t in times_utc]
    y = y.cpu().numpy()
    y_hat = y_hat.cpu().numpy()
    y_sum = y_sum.cpu().numpy() if (y_sum is not None) else None

    batch_size = y.shape[0]

    fig, axes = plt.subplots(4, 4, figsize=(8, 8))

    for i, ax in enumerate(axes.ravel()):
        if i >= batch_size:
            ax.axis("off")
            continue

        ax.plot(times_utc[i], y[i], marker=".", color="k", label=r"$y$")

        if y_sum is not None:
            ax.plot(
                times_utc[i], y_sum[i], marker=".", linestyle="--", color="k", label=r"$y_{sum}$"
            )

        if quantiles is None:
            ax.plot(times_utc[i], y_hat[i], marker=".", color="r", label=r"$\hat{y}$")
        else:
            cm = pylab.get_cmap("twilight")
            for nq, q in enumerate(quantiles):
                ax.plot(
                    times_utc[i],
                    y_hat[i, :, nq],
                    color=cm(q),
                    label=r"$\hat{y}$" + f"({q})",
                    alpha=0.7,
                )

        ax.set_title(f"{times_utc[i][0].date()}", fontsize="small")

        xticks = [t for t in times_utc[i] if t.minute == 0][::2]
        ax.set_xticks(ticks=xticks, labels=[f"{t.hour:02}" for t in xticks], rotation=90)
        ax.grid()

    axes[0, 0].legend(loc="best")

    for ax in axes[-1, :]:
        ax.set_xlabel("Time (hour of day)")

    if batch_idx is not None:
        title = f"Normed National output : batch_idx={batch_idx}"
    else:
        title = "Normed National output"
    plt.suptitle(title)
    plt.tight_layout()

    return fig



"""Base model for all PVNet submodels"""
import logging
from typing import Optional

import lightning.pytorch as pl
import torch
import torch.nn.functional as F
import wandb
from nowcasting_utils.models.loss import WeightedLosses
from nowcasting_utils.models.metrics import (
    mae_each_forecast_horizon,
    mse_each_forecast_horizon,
)
from pvnet.models.base_model import BaseModel as PVNetBaseModel
from pvnet.models.base_model import PVNetModelHubMixin
from pvnet.models.utils import (
    MetricAccumulator,
    PredAccumulator,
)
from pvnet.optimizers import AbstractOptimizer

from pvnet_summation.utils import plot_forecasts

logger = logging.getLogger(__name__)

activities = [torch.profiler.ProfilerActivity.CPU]
if torch.cuda.is_available():
    activities.append(torch.profiler.ProfilerActivity.CUDA)


class BaseModel(PVNetBaseModel):
    """Abtstract base class for PVNet summation submodels"""

    def __init__(
        self,
        model_name: str,
        model_version: Optional[str],
        optimizer: AbstractOptimizer,
        output_quantiles: Optional[list[float]] = None,
    ):
        """Abtstract base class for PVNet summation submodels.

        Args:
            model_name: Model path either locally or on huggingface.
            model_version: Model version if using huggingface. Set to None if using local.
            optimizer (AbstractOptimizer): Optimizer
            output_quantiles: A list of float (0.0, 1.0) quantiles to predict values for. If set to
                None the output is a single value.
        """
        pl.LightningModule.__init__(self)
        PVNetModelHubMixin.__init__(self)

        self.pvnet_model_name = model_name
        self.pvnet_model_version = model_version

        self.pvnet_model = PVNetBaseModel.from_pretrained(
            model_name,
            revision=model_version,
        )
        self.pvnet_model.requires_grad_(False)

        self._optimizer = optimizer

        # Model must have lr to allow tuning
        # This setting is only used when lr is tuned with callback
        self.lr = None

        self.forecast_minutes = self.pvnet_model.forecast_minutes
        self.output_quantiles = output_quantiles

        # Number of timestemps for 30 minutely data
        self.forecast_len_30 = self.forecast_minutes // 30

        self.weighted_losses = WeightedLosses(forecast_length=self.forecast_len_30)

        self._accumulated_metrics = MetricAccumulator()
        self._accumulated_y = PredAccumulator()
        self._accumulated_y_hat = PredAccumulator()
        self._accumulated_y_sum = PredAccumulator()
        self._accumulated_times = PredAccumulator()

    def predict_pvnet_batch(self, batch):
        """Use PVNet model to create predictions for batch"""
        gsp_batches = []
        for sample in batch:
            preds = self.pvnet_model(sample)
            gsp_batches += [preds]
        return torch.stack(gsp_batches)

    def sum_of_gsps(self, x):
        """Compute the sume of the GSP-level predictions"""
        if self.pvnet_model.use_quantile_regression:
            y_hat = self.pvnet_model._quantiles_to_prediction(x["pvnet_outputs"])
        else:
            y_hat = x["pvnet_outputs"]

        return (y_hat * x["effective_capacity"]).sum(dim=1)

    @property
    def pvnet_output_shape(self):
        """Return the expected shape of the PVNet outputs"""
        if self.pvnet_model.use_quantile_regression:
            return (317, self.pvnet_model.forecast_len_30, len(self.pvnet_model.output_quantiles))
        else:
            return (317, self.pvnet_model.forecast_len_30)

    def _training_accumulate_log(self, batch_idx, losses, y_hat, y, y_sum, times):
        """Internal function to accumulate training batches and log results.

        This is used when accummulating grad batches. Should make the variability in logged training
        step metrics indpendent on whether we accumulate N batches of size B or just use a larger
        batch size of N*B with no accumulaion.
        """

        losses = {k: v.detach().cpu() for k, v in losses.items()}
        y_hat = y_hat.detach().cpu()

        self._accumulated_metrics.append(losses)
        self._accumulated_y_hat.append(y_hat)
        self._accumulated_y.append(y)
        self._accumulated_y_sum.append(y_sum)
        self._accumulated_times.append(times)

        if not self.trainer.fit_loop._should_accumulate():
            losses = self._accumulated_metrics.flush()
            y_hat = self._accumulated_y_hat.flush()
            y = self._accumulated_y.flush()
            y_sum = self._accumulated_y_sum.flush()
            times = self._accumulated_times.flush()

            self.log_dict(
                losses,
                on_step=True,
                on_epoch=True,
            )

            # Number of accumulated grad batches
            grad_batch_num = (batch_idx + 1) / self.trainer.accumulate_grad_batches

            # We only create the figure every 8 log steps
            # This was reduced as it was creating figures too often
            if grad_batch_num % (8 * self.trainer.log_every_n_steps) == 0:
                fig = plot_forecasts(
                    y,
                    y_hat,
                    times,
                    batch_idx,
                    quantiles=self.output_quantiles,
                    y_sum=y_sum,
                )
                fig.savefig("latest_logged_train_batch.png")

    def training_step(self, batch, batch_idx):
        """Run training step"""

        y_hat = self.forward(batch)
        y = batch["national_targets"]
        times = batch["times"]
        y_sum = self.sum_of_gsps(batch)

        losses = self._calculate_common_losses(y, y_hat)
        losses = {f"{k}/train": v for k, v in losses.items()}

        self._training_accumulate_log(batch_idx, losses, y_hat, y, y_sum, times)

        if self.use_quantile_regression:
            opt_target = losses["quantile_loss/train"]
        else:
            opt_target = losses["MAE/train"]
        return opt_target

    def validation_step(self, batch: dict, batch_idx):
        """Run validation step"""

        y_hat = self.forward(batch)
        y = batch["national_targets"]
        times = batch["times"]
        y_sum = self.sum_of_gsps(batch)

        losses = self._calculate_common_losses(y, y_hat)
        losses.update(self._calculate_val_losses(y, y_hat))

        logged_losses = {f"{k}/val": v for k, v in losses.items()}

        # Add losses for sum of GSP predictions
        gsp_sum_losses = {
            "MSE/val": F.mse_loss(y_sum, y),
            "MAE/val": F.l1_loss(y_sum, y),
        }
        mse_each_step = mse_each_forecast_horizon(output=y_sum, target=y)
        mae_each_step = mae_each_forecast_horizon(output=y_sum, target=y)
        gsp_sum_losses.update({f"MSE_horizon/step_{i:02}": m for i, m in enumerate(mse_each_step)})
        gsp_sum_losses.update({f"MAE_horizon/step_{i:02}": m for i, m in enumerate(mae_each_step)})

        logged_losses.update({f"{k}_gsp_sum": v for k, v in gsp_sum_losses.items()})

        self.log_dict(
            logged_losses,
            on_step=False,
            on_epoch=True,
        )

        accum_batch_num = batch_idx // self.trainer.accumulate_grad_batches

        if accum_batch_num in [0, 1]:
            # Store these temporarily under self
            if not hasattr(self, "_val_y_hats"):
                self._val_y_hats = PredAccumulator()
                self._val_y = PredAccumulator()
                self._val_y_sum = PredAccumulator()
                self._val_times = PredAccumulator()

            self._val_y_hats.append(y_hat)
            self._val_y.append(y)
            self._val_y_sum.append(y_sum)
            self._val_times.append(times)

            # if batch had accumulated
            if (batch_idx + 1) % self.trainer.accumulate_grad_batches == 0:
                y_hat = self._val_y_hats.flush()
                y = self._val_y.flush()
                y_sum = self._val_y_sum.flush()
                times = self._val_times.flush()

                fig = plot_forecasts(
                    y,
                    y_hat,
                    times,
                    batch_idx,
                    quantiles=self.output_quantiles,
                    y_sum=y_sum,
                )

                self.logger.experiment.log(
                    {
                        f"val_forecast_samples/batch_idx_{accum_batch_num}": wandb.Image(fig),
                    }
                )
                del self._val_y_hats
                del self._val_y
                del self._val_y_sum
                del self._val_times

        return logged_losses

    def test_step(self, batch, batch_idx):
        """Run test step"""

        y_hat = self.forward(batch)
        y = batch["national_targets"]

        losses = self._calculate_common_losses(y, y_hat)
        losses.update(self._calculate_val_losses(y, y_hat))
        losses.update(self._calculate_test_losses(y, y_hat))
        logged_losses = {f"{k}/test": v for k, v in losses.items()}

        self.log_dict(
            logged_losses,
            on_step=False,
            on_epoch=True,
        )

        return logged_losses

    def configure_optimizers(self):
        """Configure the optimizers using learning rate found with LR finder if used"""
        if self.lr is not None:
            # Use learning rate found by learning rate finder callback
            self._optimizer.lr = self.lr
        return self._optimizer(self.parameters())



""" Data module for pytorch lightning """

import torch
from lightning.pytorch import LightningDataModule
from ocf_datapipes.load import OpenGSP
from ocf_datapipes.training.pvnet import normalize_gsp
from ocf_datapipes.utils.consts import BatchKey
from torchdata.dataloader2 import DataLoader2, MultiProcessingReadingService
from torchdata.datapipes.iter import FileLister, IterDataPipe, Zipper

# https://github.com/pytorch/pytorch/issues/973
torch.multiprocessing.set_sharing_strategy("file_system")


class GetNationalPVLive(IterDataPipe):
    """Select national output targets for given times"""

    def __init__(self, gsp_data, times_datapipe):
        """Select national output targets for given times

        Args:
            gsp_data: xarray Dataarray of the national outputs
            times_datapipe: IterDataPipe yeilding arrays of target times.
        """
        self.gsp_data = gsp_data
        self.times_datapipe = times_datapipe

    def __iter__(self):
        gsp_data = self.gsp_data
        for times in self.times_datapipe:
            national_outputs = torch.as_tensor(
                gsp_data.sel(time_utc=times.cpu().numpy().astype("datetime64[s]")).values
            )
            yield national_outputs


class GetBatchTime(IterDataPipe):
    """Extract the valid times from the concurrent sample batch"""

    def __init__(self, sample_datapipe):
        """Extract the valid times from the concurrent sample batch

        Args:
            sample_datapipe: IterDataPipe yeilding concurrent sample batches
        """
        self.sample_datapipe = sample_datapipe

    def __iter__(self):
        for sample in self.sample_datapipe:
            # Times for each GSP in the sample batch should be the same - take first
            id0 = sample[BatchKey.gsp_t0_idx]
            times = sample[BatchKey.gsp_time_utc][0, id0 + 1 :]
            yield times


class PivotDictList(IterDataPipe):
    """Convert list of dicts to dict of lists"""

    def __init__(self, source_datapipe):
        """Convert list of dicts to dict of lists

        Args:
            source_datapipe: Datapipe yielding lists of dicts
        """
        self.source_datapipe = source_datapipe

    def __iter__(self):
        for list_of_dicts in self.source_datapipe:
            keys = list_of_dicts[0].keys()
            batch_dict = {k: [d[k] for d in list_of_dicts] for k in keys}
            yield batch_dict


class DictApply(IterDataPipe):
    """Apply functions to elements of a dictionary and return processed dictionary."""

    def __init__(self, source_datapipe, **transforms):
        """Apply functions to elements of a dictionary and return processed dictionary.

        Args:
            source_datapipe: Datapipe which yields dicts
            **transforms: key-function pairs
        """
        self.source_datapipe = source_datapipe
        self.transforms = transforms

    def __iter__(self):
        for d in self.source_datapipe:
            for key, function in self.transforms.items():
                d[key] = function(d[key])
            yield d


class ZipperDict(IterDataPipe):
    """Yield samples from multiple datapipes as a dict"""

    def __init__(self, **datapipes):
        """Yield samples from multiple datapipes as a dict.

        Args:
            **datapipes: Named datapipes
        """
        self.keys = list(datapipes.keys())
        self.source_datapipes = Zipper(*[datapipes[key] for key in self.keys])

    def __iter__(self):
        for outputs in self.source_datapipes:
            yield {key: value for key, value in zip(self.keys, outputs)}  # noqa: B905


def get_capacity(batch):
    """Extract the capacity from the numpy batch"""
    return batch[BatchKey.gsp_effective_capacity_mwp]


def divide(args):
    """Divide first argument by second"""
    return args[0] / args[1]


class DataModule(LightningDataModule):
    """Datamodule for training pvnet_summation."""

    def __init__(
        self,
        batch_dir: str,
        gsp_zarr_path: str,
        batch_size=16,
        num_workers=0,
        prefetch_factor=2,
    ):
        """Datamodule for training pvnet_summation.

        Args:
            batch_dir: Path to the directory of pre-saved batches.
            gsp_zarr_path: Path to zarr file containing GSP ID 0 outputs
            batch_size: Batch size.
            num_workers: Number of workers to use in multiprocess batch loading.
            prefetch_factor: Number of data will be prefetched at the end of each worker process.
        """
        super().__init__()
        self.gsp_zarr_path = gsp_zarr_path
        self.batch_size = batch_size
        self.batch_dir = batch_dir

        self.readingservice_config = dict(
            num_workers=num_workers,
            multiprocessing_context="spawn",
            worker_prefetch_cnt=prefetch_factor,
        )

    def _get_premade_batches_datapipe(self, subdir, shuffle=False, add_filename=False):
        # Load presaved concurrent sample batches
        file_pipeline = FileLister(f"{self.batch_dir}/{subdir}", masks="*.pt", recursive=False)

        if shuffle:
            file_pipeline = file_pipeline.shuffle(buffer_size=1000)

        file_pipeline = file_pipeline.sharding_filter()

        if add_filename:
            file_pipeline, file_pipeline_copy = file_pipeline.fork(2, buffer_size=5)

        sample_pipeline = file_pipeline.map(torch.load)

        # Find national outout simultaneous to concurrent samples
        gsp_data = (
            next(iter(OpenGSP(gsp_pv_power_zarr_path=self.gsp_zarr_path).map(normalize_gsp)))
            .sel(gsp_id=0)
            .compute()
        )

        sample_pipeline, sample_pipeline_copy = sample_pipeline.fork(2, buffer_size=5)
        times_datapipe = GetBatchTime(sample_pipeline_copy)

        times_datapipe, times_datapipe_copy = times_datapipe.fork(2, buffer_size=5)
        national_targets_datapipe = GetNationalPVLive(gsp_data, times_datapipe_copy)

        times_datapipe, times_datapipe_copy = times_datapipe.fork(2, buffer_size=5)
        national_capacity_datapipe = GetNationalPVLive(
            gsp_data.effective_capacity_mwp, times_datapipe_copy
        )
        sample_pipeline, sample_pipeline_copy = sample_pipeline.fork(2, buffer_size=5)
        gsp_capacity_pipeline = sample_pipeline_copy.map(get_capacity)

        capacity_pipeline = gsp_capacity_pipeline.zip(national_capacity_datapipe).map(divide)

        # Compile the samples
        if add_filename:
            data_pipeline = ZipperDict(
                pvnet_inputs=sample_pipeline,
                effective_capacity=capacity_pipeline,
                national_targets=national_targets_datapipe,
                times=times_datapipe,
                filepath=file_pipeline_copy,
            )
        else:
            data_pipeline = ZipperDict(
                pvnet_inputs=sample_pipeline,
                effective_capacity=capacity_pipeline,
                national_targets=national_targets_datapipe,
                times=times_datapipe,
            )

        if self.batch_size is not None:
            data_pipeline = PivotDictList(data_pipeline.batch(self.batch_size))
            data_pipeline = DictApply(
                data_pipeline,
                effective_capacity=torch.stack,
                national_targets=torch.stack,
                times=torch.stack,
            )

        return data_pipeline

    def train_dataloader(self, shuffle=True, add_filename=False):
        """Construct train dataloader"""
        datapipe = self._get_premade_batches_datapipe(
            "train", shuffle=shuffle, add_filename=add_filename
        )

        rs = MultiProcessingReadingService(**self.readingservice_config)
        return DataLoader2(datapipe, reading_service=rs)

    def val_dataloader(self, shuffle=False, add_filename=False):
        """Construct val dataloader"""
        datapipe = self._get_premade_batches_datapipe(
            "val", shuffle=shuffle, add_filename=add_filename
        )
        rs = MultiProcessingReadingService(**self.readingservice_config)
        return DataLoader2(datapipe, reading_service=rs)

    def test_dataloader(self):
        """Construct test dataloader"""
        raise NotImplementedError


class PVNetPresavedDataModule(LightningDataModule):
    """Datamodule for loading pre-saved PVNet predictions to train pvnet_summation."""

    def __init__(
        self,
        batch_dir: str,
        batch_size=16,
        num_workers=0,
        prefetch_factor=2,
    ):
        """Datamodule for loading pre-saved PVNet predictions to train pvnet_summation.

        Args:
            batch_dir: Path to the directory of pre-saved batches.
            batch_size: Batch size.
            num_workers: Number of workers to use in multiprocess batch loading.
            prefetch_factor: Number of data will be prefetched at the end of each worker process.
        """
        super().__init__()
        self.batch_size = batch_size
        self.batch_dir = batch_dir

        self.readingservice_config = dict(
            num_workers=num_workers,
            multiprocessing_context="spawn",
            worker_prefetch_cnt=prefetch_factor,
        )

    def _get_premade_batches_datapipe(self, subdir, shuffle=False):
        # Load presaved concurrent sample batches
        file_pipeline = FileLister(f"{self.batch_dir}/{subdir}", masks="*.pt", recursive=False)

        if shuffle:
            file_pipeline = file_pipeline.shuffle(buffer_size=1000)

        sample_pipeline = file_pipeline.sharding_filter().map(torch.load)

        if self.batch_size is not None:
            batch_pipeline = PivotDictList(sample_pipeline.batch(self.batch_size))
            batch_pipeline = DictApply(
                batch_pipeline,
                effective_capacity=torch.stack,
                pvnet_outputs=torch.stack,
                national_targets=torch.stack,
                times=torch.stack,
            )

        return batch_pipeline

    def train_dataloader(self, shuffle=True):
        """Construct train dataloader"""
        datapipe = self._get_premade_batches_datapipe(
            "train",
            shuffle=shuffle,
        )

        rs = MultiProcessingReadingService(**self.readingservice_config)
        return DataLoader2(datapipe, reading_service=rs)

    def val_dataloader(self, shuffle=False):
        """Construct val dataloader"""
        datapipe = self._get_premade_batches_datapipe(
            "val",
            shuffle=shuffle,
        )
        rs = MultiProcessingReadingService(**self.readingservice_config)
        return DataLoader2(datapipe, reading_service=rs)

    def test_dataloader(self):
        """Construct test dataloader"""
        raise NotImplementedError
